/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// in_x.c -- X Window System input/output driver code

#include "quakedef.h"
#include "xquake.h"

Display *x_disp = NULL;
Window x_win;

qboolean mouse_available = false; // Mouse available for use
qboolean keyboard_grab_active = false;
qboolean mouse_grab_active = false;
qboolean dga_mouse_available = false;
qboolean dga_keyboard_available = false;
qboolean dga_mouse_active = false;
qboolean dga_keyboard_active = false;

int p_mouse_x, p_mouse_y;
float mouse_x, mouse_y;
int old_mouse_x, old_mouse_y;

cvar_t m_filter = {"m_filter", "0", true};

qboolean win_notify = false;
int win_width;
int win_height;
int win_x;
int win_y;

qboolean vidmode_fullscreen = false; // was vidmode_active

// if window is not the active window, don't hog as much CPU time,
// let go of the mouse, turn off sound, and restore system gamma ramps...
qboolean vid_activewindow; 
// if window is hidden, don't update screen
qboolean vid_hidden;
// if mouse entered/leaved window
qboolean vid_notifywindow;


//
// signal handler
//
/*
===============
signal_handler
===============
*/
void signal_handler(int sig)
{
	printf("Received signal %d, exiting...\n", sig);
	Sys_Quit();
}

/*
===============
InitSig

trap signals
===============
*/
void InitSig (void)
{
	signal(SIGHUP, signal_handler);
	signal(SIGINT, signal_handler);
	signal(SIGQUIT, signal_handler);
	signal(SIGILL, signal_handler);
	signal(SIGTRAP, signal_handler);
	signal(SIGIOT, signal_handler);
	signal(SIGBUS, signal_handler);
	signal(SIGFPE, signal_handler);
	signal(SIGSEGV, signal_handler);
	signal(SIGTERM, signal_handler);
}


/*
===========
XLateKey

Transform from X key symbols to Quake's symbols
===========
*/
int XLateKey (XKeyEvent *ev)
{
	int key;
	char buf[64];
	KeySym keysym;

	key = 0;

	XLookupString(ev, buf, sizeof(buf), &keysym, 0);

	switch(keysym)
	{
		case XK_KP_Page_Up:
		case XK_Page_Up:
			key = K_PGUP;
			break;

		case XK_KP_Page_Down:
		case XK_Page_Down:
			key = K_PGDN;
			break;

		case XK_KP_Home:
		case XK_Home:
			key = K_HOME;
			break;

		case XK_KP_End:
		case XK_End:
			key = K_END;
			break;

		case XK_KP_Left:
		case XK_Left:
			key = K_LEFTARROW;
			break;

		case XK_KP_Right:
		case XK_Right:
			key = K_RIGHTARROW;
			break;

		case XK_KP_Down:
		case XK_Down:
			key = K_DOWNARROW;
			break;

		case XK_KP_Up:
		case XK_Up:
			key = K_UPARROW;
			break;

		case XK_Escape:
			key = K_ESCAPE;
			break;

		case XK_KP_Enter:
		case XK_Return:
			key = K_ENTER;
			break;

		case XK_Tab:
			key = K_TAB; 
			break;

		case XK_F1: 
			key = K_F1;
			break;

		case XK_F2: 
			key = K_F2;
			break;

		case XK_F3: 
			key = K_F3;
			break;

		case XK_F4: 
			key = K_F4;
			break;

		case XK_F5:
			key = K_F5;
			break;

		case XK_F6:
			key = K_F6;
			break;

		case XK_F7:
			key = K_F7;
			break;

		case XK_F8: 
			key = K_F8;
			break;

		case XK_F9:
			key = K_F9;
			break;

		case XK_F10:
			key = K_F10;
			break;

		case XK_F11:
			key = K_F11;
			break;

		case XK_F12:
			key = K_F12;
			break;

		case XK_BackSpace: 
			key = K_BACKSPACE; 
			break;

		case XK_KP_Delete:
		case XK_Delete: 
			key = K_DEL; 
			break;

		case XK_Pause:
			key = K_PAUSE;
			break;

		case XK_Shift_L:
		case XK_Shift_R:
			key = K_SHIFT;
			break;

		case XK_Execute:
		case XK_Control_L:
		case XK_Control_R:
			key = K_CTRL;
			break;

		case XK_Alt_L:
		case XK_Meta_L:
		case XK_Alt_R:
		case XK_Meta_R:
			key = K_ALT;
			break;

		case XK_KP_Begin: 
			key = K_AUX30;
			break; 

		case XK_Insert:
		case XK_KP_Insert: 
			key = K_INS; 
			break;

		case XK_KP_Multiply: 
			key = '*'; 
			break;

		case XK_KP_Add: 
			key = '+'; 
			break;

		case XK_KP_Subtract: 
			key = '-'; 
			break;

		case XK_KP_Divide: 
			key = '/'; 
			break;

		case XK_section:
			key = '~';
			break;

		case XK_Caps_Lock: 
			key = K_CAPSLOCK; 
			break;

		case XK_Num_Lock: 
			key = KP_NUMLOCK; 
			break;

		default:
			key = *(byte *)buf;
			if (key >= 'A' && key <= 'Z')
				key = key - 'A' + 'a';
			break;
	} 
	return key;
}


/*
===========
CreateNullCursor

makes a null cursor
===========
*/
static Cursor CreateNullCursor (void)
{
	Pixmap cursormask;
	XGCValues xgc;
	GC gc;
	XColor dummycolour;
	Cursor cursor;

	cursormask = XCreatePixmap(x_disp, x_win, 1, 1, 1 /* depth */ );
	xgc.function = GXclear;
	gc = XCreateGC(x_disp, cursormask, GCFunction, &xgc);
	XFillRectangle(x_disp, cursormask, gc, 0, 0, 1, 1);
	dummycolour.pixel = 0;
	dummycolour.flags = 0;
	cursor = XCreatePixmapCursor(x_disp, cursormask, cursormask, &dummycolour, &dummycolour, 0, 0);
	XFreePixmap(x_disp, cursormask);
	XFreeGC(x_disp, gc);

	return cursor;
}


/*
===========
IN_ActivateDGAMouse
===========
*/
static void IN_ActivateDGAMouse (void)
{
	int DGAflags;

	if (dga_mouse_available && !dga_mouse_active)
	{
		DGAflags = XF86DGADirectMouse;
		if (dga_keyboard_active)
		{
			XF86DGADirectVideo(x_disp, DefaultScreen(x_disp), 0);
			DGAflags |= XF86DGADirectKeyb;
		}

		XF86DGADirectVideo(x_disp, DefaultScreen(x_disp), DGAflags);
		dga_mouse_active = true;
	}
}

/*
===========
IN_DeactivateDGAMouse
===========
*/
static void IN_DeactivateDGAMouse (void)
{
	int DGAflags;

	if (dga_mouse_available && dga_mouse_active)
	{
		DGAflags = 0;
		if (dga_keyboard_active)
		{
			XF86DGADirectVideo(x_disp, DefaultScreen(x_disp), 0);
			DGAflags |= XF86DGADirectKeyb;
		}

		XF86DGADirectVideo(x_disp, DefaultScreen(x_disp), DGAflags);
		dga_mouse_active = false;
	}
}

/*
===========
IN_GrabMouse
===========
*/
void IN_GrabMouse (void)
{
	if (mouse_available && !mouse_grab_active && x_win)
	{
		XWindowAttributes attribs_1;
		XSetWindowAttributes attribs_2;

		XGetWindowAttributes(x_disp, x_win, &attribs_1);
		attribs_2.event_mask = attribs_1.your_event_mask | KEY_MASK | MOUSE_MASK;
		XChangeWindowAttributes(x_disp, x_win, CWEventMask, &attribs_2);

		// hide cursor
		XDefineCursor(x_disp, x_win, CreateNullCursor());
		// grab pointer
		XGrabPointer(x_disp, x_win, True, 0, GrabModeAsync, GrabModeAsync, x_win, None, CurrentTime);

		if (dga_mouse_available)
		{
			IN_ActivateDGAMouse();
		}
		else
		{
			XWarpPointer(x_disp, None, x_win, 0, 0, 0, 0, vid.width / 2, vid.height / 2);
		}

		mouse_x = mouse_y = 0;
		mouse_grab_active = true;
	}
}

/*
===========
IN_UngrabMouse
===========
*/
void IN_UngrabMouse (void)
{
	if (mouse_grab_active)
	{
		if (dga_mouse_active)
		{
			IN_DeactivateDGAMouse();
			XWarpPointer(x_disp, None, x_win, 0, 0, 0, 0, 0, 0);
		}

		// ungrab pointer
		XUngrabPointer(x_disp, CurrentTime);
		// show cursor
		if (x_win)
			XUndefineCursor(x_disp, x_win);

		mouse_grab_active = false;
	}
}


/*
===========
IN_ActivateDGAKeyboard
===========
*/
static void IN_ActivateDGAKeyboard (void)
{
	int DGAflags;

	if (dga_keyboard_available && !dga_keyboard_active)
	{
		DGAflags = XF86DGADirectKeyb;
		if (dga_mouse_active)
		{
			XF86DGADirectVideo(x_disp, DefaultScreen(x_disp), 0);
			DGAflags |= XF86DGADirectMouse;
		}

		XF86DGADirectVideo(x_disp, DefaultScreen(x_disp), DGAflags);
		dga_keyboard_active = true;
	}
}

/*
===========
IN_DeactivateDGAKeyboard
===========
*/
static void IN_DeactivateDGAKeyboard (void)
{
	int DGAflags;

	if (dga_keyboard_available && dga_keyboard_active)
	{
		DGAflags = 0;
		if (dga_mouse_active)
		{
			XF86DGADirectVideo(x_disp, DefaultScreen(x_disp), 0);
			DGAflags |= XF86DGADirectMouse;
		}

		XF86DGADirectVideo(x_disp, DefaultScreen(x_disp), DGAflags);
		dga_keyboard_active = false;
	}
}

/*
===========
IN_GrabKeyboard
===========
*/
void IN_GrabKeyboard (void)
{
	if (!keyboard_grab_active && x_win)
	{
		// grab keyboard
		XGrabKeyboard(x_disp, x_win, False, GrabModeAsync, GrabModeAsync, CurrentTime);

		if (dga_keyboard_available)
		{
			if (!vidmode_fullscreen && (vid.width < 400 || vid.height < 300))
				Con_Warning ("Running low-res windowed mode, XFree86 DGA Keyboard disabled\n");
			else
				IN_ActivateDGAKeyboard();
		}

		keyboard_grab_active = true;
	}
}

/*
===========
IN_UngrabKeyboard
===========
*/
void IN_UngrabKeyboard (void)
{
	if (keyboard_grab_active)
	{
		if (dga_keyboard_active)
		{
			IN_DeactivateDGAKeyboard();
		}

		// ungrab keyboard
		XUngrabKeyboard(x_disp, CurrentTime);

		keyboard_grab_active = false;
	}
}


/*
===========
Force_CenterView_f
===========
*/
void Force_CenterView_f (void)
{
	cl.viewangles[PITCH] = 0;
}

/*
===========
IN_Init
===========
*/
void IN_Init (void)
{
	int MajorVersion, MinorVersion;
	qboolean DGA = false;

	Cvar_RegisterVariable (&m_filter);

	Cmd_AddCommand ("force_centerview", Force_CenterView_f);

	keyboard_grab_active = false;
	dga_keyboard_active = false;

	mouse_grab_active = false;
	dga_mouse_active = false;

	if (COM_CheckParm ("-nomouse"))
		mouse_available = false;
	else
		mouse_available = true;

	if (x_disp == NULL)
		Sys_Error ("IN_Init: x_disp not initialised before input...\n");

	MajorVersion = MinorVersion = 0;
	DGA = XF86DGAQueryVersion(x_disp, &MajorVersion, &MinorVersion);

	if(COM_CheckParm("-nodga"))
	{
		Con_Warning ("XFree86 DGA disabled at command line\n");
	}
	else if (DGA) 
	{
		Con_Printf ("XFree86 DGA Version %d.%d found\n", MajorVersion, MinorVersion);

		if (COM_CheckParm("-nodgamouse"))
			Con_Warning ("XFree86 DGA Mouse disabled at command line\n");
		else
			dga_mouse_available = true;

		if (COM_CheckParm("-nodgakeyb"))
			Con_Warning ("XFree86 DGA Keyboard disabled at command line\n");
		else
			dga_keyboard_available = true;
	}
	else
	{
		Con_Warning ("XFree86 DGA not supported\n");
	}

	IN_GrabMouse(); // grab mouse first!
	IN_GrabKeyboard();
}

/*
===========
IN_Shutdown
===========
*/
void IN_Shutdown (void)
{
	IN_UngrabMouse();
	IN_UngrabKeyboard();

	mouse_available = false;
}

/*
===========
IN_Commands
===========
*/
void IN_Commands (void)
{
}

/*
===========
IN_MouseMove
===========
*/
void IN_MouseMove (usercmd_t *cmd)
{
	int	tx, ty;

	if (!mouse_available)
		return;

	tx = mouse_x;
	ty = mouse_y;

	if (m_filter.value)
	{
		mouse_x = (tx + old_mouse_x) * 0.5;
		mouse_y = (ty + old_mouse_y) * 0.5;
	}

	old_mouse_x = tx;
	old_mouse_y = ty;

	mouse_x *= sensitivity.value;
	mouse_y *= sensitivity.value;

// add mouse X/Y movement to cmd
	if ( (in_strafe.state & 1) || (lookstrafe.value && (in_mlook.state & 1) ))
		cmd->sidemove += m_side.value * mouse_x;
	else
		cl.viewangles[YAW] -= m_yaw.value * mouse_x;
	
	if (in_mlook.state & 1)
		V_StopPitchDrift ();
		
	if ( (in_mlook.state & 1) && !(in_strafe.state & 1))
	{
		cl.viewangles[PITCH] += m_pitch.value * mouse_y;
		if (cl.viewangles[PITCH] > 80)
			cl.viewangles[PITCH] = 80;
		if (cl.viewangles[PITCH] < -70)
			cl.viewangles[PITCH] = -70;
	}
	else
	{
		if ((in_strafe.state & 1) && noclip_anglehack)
			cmd->upmove -= m_forward.value * mouse_y;
		else
			cmd->forwardmove -= m_forward.value * mouse_y;
	}
	mouse_x = mouse_y = 0;
}

/*
===========
IN_Move
===========
*/
void IN_Move (usercmd_t *cmd)
{
	if (vid_activewindow && !vid_hidden)
		IN_MouseMove(cmd);
}

